unit KADriveLetter;

interface

uses
  Windows, SysUtils, Classes;

type
  ActionType=(DoAssign,DoRemove,DoChange);
  TKADriveLetter = class(TComponent)
  private
    FActivate  : Boolean;
    FOldLetter : String;
    FPartition : String;
    FNewLetter : String;
    FDrives    : TStringList;
    FVolumes   : TStringList;
    FAction    : ActionType;
    procedure SetAction(const Value: ActionType);
    procedure SetActivate(const Value: Boolean);
    procedure SetNewLetter(const Value: String);
    procedure SetOldLetter(const Value: String);
    procedure SetPartition(const Value: String);
    procedure SetDrives(const Value: TStringList);
    procedure SetVolumes(const Value: TStringList);
    { Private declarations }
  protected
    { Protected declarations }
    Function  DriveExists(Drive : String):Boolean;
    Function  CorrectDriveLetter(Drive : String):String;
    Procedure EnumDrives;
    Procedure EnumVolumes;
    Function  GetDriveVolumeName(Drive : String):String;
    Function  RemoveDriveLetter(Drive : String):Boolean;
    Function  AssignDriveLetter(Drive : String; NTDevice : String):Boolean;
  public
    { Public declarations }
    Constructor Create(AOwner:TComponent); Override;
    Destructor  Destroy; Override;
  published
    { Published declarations }
    Property Action    : ActionType  read FAction    write SetAction;
    Property OldDrive  : String      read FOldLetter write SetOldLetter;
    Property NewDrive  : String      read FNewLetter write SetNewLetter;
    Property Partition : String      read FPartition write SetPartition;
    Property Drives    : TStringList read FDrives    write SetDrives;
    Property Volumes   : TStringList read FVolumes   write SetVolumes;
    Property Activate  : Boolean     read FActivate  write SetActivate;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('KA', [TKADriveLetter]);
end;

{ TKADriveLetter }
Function TKADriveLetter.DriveExists(Drive : String):Boolean;
Begin
  Result := GetDriveType(PChar(Drive+'\')) <> 1;
End;

Function TKADriveLetter.CorrectDriveLetter(Drive : String):String;
Var
  L : Integer;
Begin
  Result := Drive;
  L      := Length(Result);
  if L > 3 Then System.Delete(Result,1,2);
  L := 2;
  if Result[L] <> ':' Then
     Begin
       Result := '';
       Exit;
     End;
  if (NOT (Result[1] in ['A'..'Z'])) AND (NOT (Result[1] in ['a'..'z'])) Then
     Begin
       Result := '';
       Exit;
     End;
End;

Procedure TKADriveLetter.EnumVolumes;
Type
  TFVC=Procedure(FH:THandle);StdCall;
  TFFV=Function(lpszVolumeName:PChar;cchBufferLength:DWORD):THANDLE;StdCall;
  TFNV=Function(hFindVolume:THANDLE;lpszVolumeName:PChar;cchBufferLength:DWORD):BOOL;StdCall;
Var
 FH                : THandle;
 VolumeName        : Array[0..1000] Of Char;
 FindFirstVolume   : TFFV;
 FindNextVolume    : TFNV;
 FindVolumeClose   : TFVC;
 LibHandle         : THandle;
Begin
 FVolumes.Clear;
 LibHandle := LoadLibrary('Kernel32.dll');
 if LibHandle > 0 Then
    Begin
       @FindFirstVolume := GetProcAddress(LibHandle,'FindFirstVolumeA');
       @FindNextVolume  := GetProcAddress(LibHandle,'FindNextVolumeA');
       @FindVolumeClose := GetProcAddress(LibHandle,'FindVolumeClose');
       If  (@FindFirstVolume <> Nil)
       And (@FindNextVolume <> Nil)
       And (@FindVolumeClose <> Nil) Then
           Begin
             FH := FindFirstVolume(VolumeName,MAX_PATH);
             if FH <> INVALID_HANDLE_VALUE Then
                Begin
                  FVolumes.Add(VolumeName);
                  While FindNextVolume(FH,VolumeName,MAX_PATH) Do
                    Begin
                      FVolumes.Add(VolumeName);
                    End;
                  FindVolumeClose(FH);
               End;
           End;
       FreeLibrary(LibHandle);
    End;
End;

Procedure TKADriveLetter.EnumDrives;
Var
  X                                : Integer;
  Drive                            : String;
  UniqueVolumeName                 : Array[0..1000] of Char;
Begin
  FDrives.Clear;
  For X := Ord('A') To Ord('Z') do
      Begin
        Drive := CHR(X)+':';
        if GetDriveType(PChar(Drive+'\')) <> 1 Then
           Begin
              QueryDosDevice(PChar(Drive),UniqueVolumeName,MAX_PATH);
              FDrives.Add(Drive+'='+UniqueVolumeName);
           End;
      End;
End;

Function TKADriveLetter.GetDriveVolumeName(Drive : String):String;
Var
  I : Integer;
Begin
  Result := '';
  I := Drives.IndexOfName(Drive);
  if I <> -1 Then Result := Drives.Values[Drives.Names[I]];
End;

Function TKADriveLetter.RemoveDriveLetter(Drive : String):Boolean;
Type
 TDVMP = Function (VolumeMountPoint : PChar):Boolean;StdCall;
Var
  LibHandle              : THandle;
  DeleteVolumeMountPoint : TDVMP;
Begin
  Result    := False;
  if CorrectDriveLetter(Drive) = '' Then Exit;
  LibHandle := LoadLibrary('Kernel32.dll');
  if LibHandle > 0 Then
     Begin
        @DeleteVolumeMountPoint := GetProcAddress(LibHandle,'DeleteVolumeMountPointA');
        if @DeleteVolumeMountPoint <> Nil Then
           Begin
             Result := DeleteVolumeMountPoint(PChar(Drive+'\'));
           End;
        FreeLibrary(LibHandle);
     End;
End;

Function TKADriveLetter.AssignDriveLetter(Drive : String; NTDevice : String):Boolean;
Type
  TGVNFMP = Function(Drive:PChar; UniqueVolumeName : PChar; BufSize : DWORD):Boolean;StdCall;
  TSVMP   = Function(Drive:PChar; UniqueVolumeName : PChar):Boolean;StdCall;
Var
  LibHandle                        : THandle;
  GetVolumeNameForVolumeMountPoint : TGVNFMP;
  SetVolumeMountPoint              : TSVMP;
  UniqueVolumeName                 : Array[0..1000] of Char;
Begin
  if CorrectDriveLetter(Drive) = '' Then Exit;
  Result := DefineDosDevice(DDD_RAW_TARGET_PATH, PChar(Drive), PChar(NTDevice));
  If Result Then
     Begin
       LibHandle := LoadLibrary('Kernel32.dll');
       if LibHandle > 0 Then
          Begin
            @GetVolumeNameForVolumeMountPoint := GetProcAddress(LibHandle,'GetVolumeNameForVolumeMountPointA');
            if @GetVolumeNameForVolumeMountPoint <> Nil Then
               Begin
                  if (NOT GetVolumeNameForVolumeMountPoint(PChar(Drive+'\'), UniqueVolumeName, MAX_PATH)) Then
                     Begin
                       UniqueVolumeName[0] := #0;
                     End;
                  Result := DefineDosDevice(DDD_RAW_TARGET_PATH OR DDD_REMOVE_DEFINITION OR DDD_EXACT_MATCH_ON_REMOVE, PChar(Drive), PChar(NTDevice));
                  if Result Then
                     Begin
                       @SetVolumeMountPoint := GetProcAddress(LibHandle,'SetVolumeMountPointA');
                       if @SetVolumeMountPoint <> Nil Then
                          Begin
                            Result := SetVolumeMountPoint(PChar(Drive+'\'),UniqueVolumeName);
                          End
                       Else
                          Begin
                            Result := False;
                          End;
                     End;
               End
            Else
               Begin
                 Result := False;
               End;
          End
       Else
          Begin
            Result := False;
          End;
     End
End;

constructor TKADriveLetter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAction := DoAssign;
  FDrives  := TStringList.Create;
  FVolumes := TStringList.Create;
  EnumDrives;
  EnumVolumes;
end;

destructor TKADriveLetter.Destroy;
begin
  FDrives.Free;
  FVolumes.Free;
  inherited;
end;


procedure TKADriveLetter.SetAction(const Value: ActionType);
begin
  FAction := Value;
end;

procedure TKADriveLetter.SetActivate(const Value: Boolean);
begin
  if csLoading in ComponentState Then Exit;
  FActivate := Value;
  if FActivate Then
     Begin
      if FAction=DoAssign Then
         Begin
           FActivate := AssignDriveLetter(FNewLetter,FPartition);
         End
      Else
      if FAction=DoRemove Then
         Begin
           FActivate := RemoveDriveLetter(FOldLetter);
         End
      Else
      if FAction=DoChange Then
         Begin
           FPartition := GetDriveVolumeName(FOldLetter);
           FActivate  := RemoveDriveLetter(FOldLetter);
           if FActivate Then
              FActivate := AssignDriveLetter(FNewLetter,FPartition);
         End;
      EnumDrives;
      EnumVolumes;
     End;
end;

procedure TKADriveLetter.SetNewLetter(const Value: String);
begin
  if (CorrectDriveLetter(Value) <> '') And (NOT DriveExists(Value)) Then
     FNewLetter := Value;
end;

procedure TKADriveLetter.SetOldLetter(const Value: String);
begin
  if (CorrectDriveLetter(Value) <> '') And (DriveExists(Value)) Then
     FOldLetter := Value;
end;

procedure TKADriveLetter.SetPartition(const Value: String);
begin
  FPartition := Value;
end;


procedure TKADriveLetter.SetDrives(const Value: TStringList);
begin
  //****************************************************************** READ ONLY
end;

procedure TKADriveLetter.SetVolumes(const Value: TStringList);
begin
  //****************************************************************** READ ONLY
end;

end.
